=begin

Name: Trigram
Author: John F. McGowan, Ph.D.
Contact: jmcgowan11@earthlink.net
Web: http://www.jmcgowan.com
Copyright (C) 2009, John F. McGowan

Description: Trigram counts the ngrams in a training set of text files and computes
the correlation coefficient between the trigram and a topic (e.g. "rockets") over the training set.
An n-gram is a sequence of n words in a file.  A unigram is just a single word (e.g. "rocket").  A
bigram is two words in sequence (e.g. "rocket exploded").  A trigram is three words in sequence
(e.g. "rocket exploded violently").  A topic such as "rocket" refers to the meaning, the semantics, of 
a web page or document.  An ngram such as "rocket" or "rocket scientist" can appear in a web page or
document that has nothing to do with the topic.  For exmaple, "the rocket scientists on Wall Street
wrecked the global economy using the Gaussian copula model." in an article on mortgage backed
securities.  

The correlation coefficient ranges from -1.0 to 1.0.  1.0 means a perfect correlation in the training
set.  The ngram is always present in documents about the topic and never present in documents not
about the topic.  0.0 means no correlation.  The presence or absence of the ngram tells nothing about
the topic.  For example, common words such as "the" or "is" are found in almost all documents.
-1.0 means the ngram is perfectly negatively correlated with the topic.  If the ngram is present, the
document is not about the topic.

Statistical models of the frequency of n-grams are used in speech recognition and machine translation.  
They can also be used in web or other document search applications to locate target web pages or 
documents by topic (e.g. rocketry).  One needs to find statistical patterns of words that are highly
correlated with the search topic.  Web pages or other documents can be hand labeled or one can use
keyword tags, meta tags, and so forth to identify the topic(s) of the document for training purposes.

At present the trigram script is a simple demonstration of the concept.  A real search engine would 
use more sophisticated statistical models.

Files: 	trigram.rb  # main script
		train.txt   # demo list of training files
		triout.txt  # sample output file using the sample training files
		
		Training File	Topic  (in train.txt)
		test.txt test
		test2.txt test
		test3.txt rocket
		test4.txt rocket
		test5.txt rocket

Platform:  Developed and tested (briefly) on Windows XP Service Pack 2 (2.6 GHz) Sony Vaio 
with Ruby 1.86 installed.		
		
About the Author

John F. McGowan, Ph.D. is a software developer, research scientist, and consultant. He works 
primarily in the area of complex algorithms that embody advanced mathematical and 
logical concepts, including speech recognition and video compression technologies. He has many 
years of experience developing software in Visual Basic, C++, and many other programming languages 
and environments. He has a Ph.D. in Physics from the University of Illinois at Urbana-Champaign 
and a B.S. in Physics from the California Institute of Technology (Caltech). 
He can be reached at jmcgowan11@earthlink.net.

=end

class Trigram
	def initialize(file_name)
		@file = file_name
		@tokens = []
		@unigrams = {}
		@bigrams = {}
		@trigrams = {}
		@allgrams = {}
		
		puts "reading #{file_name}"
		@lines = IO.readlines(file_name)
		puts "read #{@lines.length} lines "
		for i in 0..@lines.length-1
			@tokens = @tokens + @lines[i].split
			# puts "\nLine #{i}"
			#puts @tokens
		end

		for i in 0..@tokens.length-1
			puts "checking #{@tokens[i]}"

			if @tokens[i][-1].chr == "." || @tokens[i][-1].chr == "," || @tokens[i][-1].chr == "!" || @tokens[i][-1].chr == "?"
				puts "removing trailing ."
				@tokens[i] = @tokens[i].chop
			end
			@tokens[i] = @tokens[i].upcase  # don't worry about case for now
		end


		@tokens.each { |tok| 
				
				# puts "checking #{tok} "
				if @unigrams[tok] 	
					@unigrams[tok] = @unigrams[tok] + 1
                                else
					@unigrams[tok] = 1
				end
		}  # collect histogram of 1-grams

		for i in 0..@tokens.length-2
			if @bigrams[@tokens[i] + "." + @tokens[i+1] ]
				@bigrams[@tokens[i] + "." + @tokens[i+1]] = @bigrams[@tokens[i] + "." + @tokens[i+1]] + 1
			else
				@bigrams[@tokens[i] + "." + @tokens[i+1]] = 1
			end
		end  # collect histogram of bigrams

		for i in 0..@tokens.length-3
			trig = @tokens[i] + "." + @tokens[i+1] + "." + @tokens[i+2]
			if @trigrams[trig]
				@trigrams[trig] = @trigrams[trig] + 1
			else
				@trigrams[trig] = 1
			end
		end  # collect histogram of trigrams


		@allgrams = @unigrams.merge(@bigrams)
		@allgrams = @allgrams.merge(@trigrams)

	end  # initialize

	def show
		puts "1-grams #{@file}"
		puts @unigrams
		puts "\n2-grams #{@file}"
		puts @bigrams
		puts "\n3-grams #{@file}"
		puts @trigrams

		puts "\nAll N Grams #{@file}"
		puts @allgrams
	end

	def unigrams
		return @unigrams.keys
	end

	def bigrams
		return @bigrams.keys
	end

	def trigrams
		return @trigrams.keys
	end

	def allgrams
		return @allgrams.keys
	end

	def has_ngram?(ngram)
		return @allgrams.has_key?(ngram)
	end

end


class Trainer
	def initialize(file_name)
		@file = file_name  # file contains a list of file <space> topic
		@training_file = []
		@topic = []
		@topic_file = {}
		@topic_average = {}
		@StatData = {}
		@number_files = {}
		@allmatches = {}
		@allfilem = {}
		@allngrams = []  
		@correlation = {}  # indexed by n-gram and topic

		@lines = IO.readlines(file_name)

		@lines.each { |line|
			@dummy = line.split
			@training_file.push @dummy[0]  # text file
			@topic.push @dummy[1]          # topic
			@topic_file[@dummy[0]] = @dummy[1]  # quick lookup of file topic
			@StatData[@dummy[0]] = Trigram.new(@dummy[0])
			@StatData[@dummy[0]].show
			@local_tokens = []

			if @number_files[@dummy[1]]  # number of files on topic
				@number_files[@dummy[1]] = @number_files[@dummy[1]] + 1
			else
				@number_files[@dummy[1]] = 1.0
			end 

# count how many files contain a unigram
			@matches = @StatData[@dummy[0]].unigrams  # list of unigrams for file

			@matches.each { |match|
				key = match + "." + @dummy[1]  # combine keyword and topic
				if @allmatches[key] 
					@allmatches[key] = @allmatches[key] + 1.0
				else
					@allmatches[key] = 1.0
				end

				if @allfilem[match]
					@allfilem[match] = @allfilem[match] + 1.0
				else
					@allfilem[match] = 1.0
				end

			} # end of block

# count how many files contain a bigram
			@matches = @StatData[@dummy[0]].bigrams  # list of bigrams for file

			@matches.each { |match|   # match is the bigram
				key = match + "." + @dummy[1]  # combine keyword and topic
				if @allmatches[key] 
					@allmatches[key] = @allmatches[key] + 1.0
				else
					@allmatches[key] = 1.0
				end

				if @allfilem[match]
					@allfilem[match] = @allfilem[match] + 1.0
				else
					@allfilem[match] = 1.0
				end

			}  # end of block

# count how many files contain a trigram
			@matches = @StatData[@dummy[0]].trigrams  # list of trigrams for file

			@matches.each { |match|   # match is the trigram
				key = match + "." + @dummy[1]  # combine keyword and topic
				if @allmatches[key] 
					@allmatches[key] = @allmatches[key] + 1.0
				else
					@allmatches[key] = 1.0
				end

				if @allfilem[match]
					@allfilem[match] = @allfilem[match] + 1.0
				else
					@allfilem[match] = 1.0
				end

			}  # end of block
			
			@allngrams = @allngrams + @StatData[@dummy[0]].allgrams  # build list of all ngrams in all files				
				
		} # loop over files in training set				
		@number_files["ALL"] = @lines.length.to_f
		
		# @allmatches = @allmatches / @number_files

		@allmatches.each_key { |key|
				@local_tokens = key.split(".")
				@allmatches[key] = @allmatches[key] / @number_files[@local_tokens[-1]]
		}	

		@allfilem.each_key { |key|
				@allfilem[key] = @allfilem[key] / @number_files["ALL"]
		}       # proportion of training files with ngram.topic compound key	

		@topic.each { |top|
			@topic_average[top] = @number_files[top]/@number_files["ALL"]
		}	# proportion of training files with topic

		@allngrams.each { |ngram|  # loop over ngrams in training set
			@topic.each { |top|  # loop over topics in training set

				puts "computing correlation coefficient for #{ngram} AND TOPIC: #{top}"

				compound_key = ngram + top
				
				@correlation[compound_key] = 0.0  # start at zero
				covariance = 0.0
				variance_ngram = 0.0
				variance_topic = 0.0

				# @correlation[ngram.topic] = covariance/Math.sqrt(variance_ngram*variance_topic)		

	
				@training_file.each { |file|
					if @topic_file[file] == top
						delta_top = (1.0 - @topic_average[top])
					else
						delta_top = (0.0 - @topic_average[top])
					end

					variance_topic = variance_topic + delta_top**2

					if @StatData[file].has_ngram?(ngram)
						delta_ngram = (1.0 - @allfilem[ngram])
					else
						delta_ngram = (0.0 - @allfilem[ngram])
					end
					variance_ngram = variance_ngram + delta_ngram**2

					covariance = covariance + delta_top*delta_ngram

				}	# loop over files in training set

				@correlation[compound_key] = covariance/ Math.sqrt(variance_ngram*variance_topic)
			}
		}

	end

	def show
		puts "\nTraining Files"
		puts @training_file
		puts "\nTopics"
		puts @topic

		puts @StatData

		puts "\nUnigram Match Statistics"
		sorted = @allmatches.sort {|a,b| a[1]<=>b[1]}  # sort result by value (not key) 
		sorted.each { |match| puts match }
		puts "\nUnigram Statistics over All Files"
		sorted = @allfilem.sort {|a,b| a[1] <=> b[1]}  # sort by value
		sorted.each { |match| puts match }

		puts "\nTopic Average"
		puts @topic_average


		puts "\nCorrelations"
		sorted = @correlation.sort {|a,b| a[1] <=> b[1]}  # sort by value
		sorted.each { |match| puts match }


	end 
end

# simple demo

#t = Trigram.new("test.txt")

#t.show

set = Trainer.new("train.txt")

set.show
 